/*

	Ptahhotep's Team Balancer (PTB)
	Version 1.6

	DATE: May 11, 2002
	FILE: ptb.sma
	AUTHOR: Ptahhotep (ptahhotep@planethalflife.com)(Ported to AMX By: Tcquest78)
	CONTRIBUTIONS: 
	o NtroP started some new team switching code for PTB 1.5,
	  choosing the switch/transfer which balances team strengths best.
	TESTERS: all the great guys at http://www.concarne.org

	Thanks to all those who reported bugs and sent in suggestions!
	Special thanks go to Innocence from http://www.seth.dk, who maintains
	a small PTB FAQ.

	Comments On Version 1.6
	-----------------------
	o This version adds another 8 new configuration options to PTB.
	o Setting the new max team size setting to a number smaller than half
	  the number of maximum players on the server effectively forces the
	  remaining players into spectator mode.
	o I put some real work into the amx_ptb command. Please check out
	  the "amx_ptb" and "amx_ptb status" commands from the CS console.
	o PTB now limits the frequency of switches of individual players,
	  which should help getting rid of complaints about frequent switching.
	o PTB 1.6 uses the AdminMod playercount() function to determine
	  the current number of players in the game. The number returned by
	  that function is not always correct. Please keep that in mind when
	  using the new PTB_SWITCHMIN and PTB_LIMITMIN settings. You can
	  monitor the value returned by playerinfo() by typing amx_ptb in
	  the console. I also tried to track the number of connections and
	  disconnections, but that gives even worse results, since AdminMod
	  doesn't seem to generate either enough disconnect or more than
	  enough connect events.
	o The new commands "amx_ptb save" and "amx_ptb load" let you
	  save and restore changes in the settings made from the console.
	  That way you can experiment with the PTB settings in-game and
	  make them permanent without editing any files by hand.
	o The new PTB_SUPERRATING setting adds a new winning team condition,
	  that effectively limits the maximum allowed team strength for the
	  losing team. Before version 1.6 PTB allowed a team to be as strong
	  as it gets, as long as it didn't have either a winning streak or
	  a score advantage or both. See the Details section for details.
	o The new commands "amx_ptb list" and "amx_ptb help" both give
	  you a brief overview of available PTB commands.

	Comments On Version 1.5
	-----------------------
	This verion of PTB is tested with AMX 0.7.3, Metamod 1.13.1
	and AMX-LogD(dJeyls's version ...).
	This version improves PTB in a lot of ways:
	o Most important is the new team switching code, which looks
	  for the best possible switch to execute. If there is no switch
	  that makes any difference in team balance, no switching happens.
	  Thanks go to NtroP for pushing me to finish his team
	  switching code for version 1.5 of PTB.
	o Also new is an improved team strength calculation method,
	  providing better indication of team strength. A detailed
	  description is given in the Details section.
	o Players will not be allowed to choose their team in the first
	  3 rounds, instead they will be auto-joined.
	o PTB will now try to keep the differences in team size below 3
	  when auto-joining players.
	o Forcing a switch (switching alive) is now configurable. You
	  can set the number of unsuccessful tries, after which PTB
	  does a switch of living players (killing them in the process).
	o Default settings for wtj auto-joining and kicking are much
	  more strict now. Try wtjing a second time and PTB will auto-
	  join you, try it a third time and you get kicked.
	o There is now an option to disable typesays completely.
	o All the things above are configurable through "vault.ini" -
	  another new feature of PTB 1.5.
	o Finally some bugs were fixed and some additional changes might
	  have happened that I forgot to note.
	See the latest changes notes below and the new section about
	vault.ini for more details.

	Introduction
	------------
	PTB is an AMX plugin (http://amx.djeyl.net).
	LogD (http://djeyl.net/ms-modules.php) is required to run PTB.
	PTB is a Counter-Strike (CS) specific plugin.

	To run PTB with its default settings, simply compile it
	for your operating system, include it into plugin.ini
	and change the map on your CS server.

	PTB detects situations, in which one team rules over the other.
	It looks at the current team scores, current team win streaks
	and the current team ratings and decides on actions with the
	help of these values.
	
	Additionally PTB completely controls team joining actions of
	players and prevents unbalancing team joins. PTB auto-joins
	and even kicks obvious winning team joiners.

	Features
	--------
	o team strength evaluation based on scores, streaks, kills, deaths and size
	o search for best balancing switch or transfer
	o automatic dead only team switches
	o automatic dead only team transfers
	o announcement of team balance status
	o display of WTJ attempts
	o prevention, auto-join and kick for WTJing
	o "safe auto-join" - overriden auto-join function, that is always
	  choosing the right team and prevents joining of the same or
	  winning or bigger team, using auto-team never kills a player,
	  if he can't switch teams, and never gives WTJ counts
	o controllable with 33 console commands
	o configurable with 26 configuration options
	o configurable through vault.ini
	o much more ...

	Details
	-------
	The most important thing PTB evaluates is team strength. PTB calculates
	a single number as an indicator of team strength.

	In PTB 1.5 the team strength calculation changed. In earlier versions,
	team strength was calculated as the sum of total kills per team over the
	sum of total deaths per team times the number of players per team
	(strength = kills/deaths * players). The method didn't always work well,
	since a very good player (20:1) and a very bad player (1:20) made no
	difference to 2 mediocre players (5:5 and 6:6).
	
	The new method calculates a separate kills/deaths ratio per player, sums
	the ratios up per team and multiplies them with team size
	(strength = players * (sum of kills/deaths per player)).

	 Example:
	 CT team: A (6:5), B (2:3)
	 T team: C (1:1), D (4:3), E (1:4)
	 CT strength: 2 * (6/5 + 2/3) = 3.73
	 T strength: 3 * (1/1 + 4/3 + 1/4) = 7.75

	To compare the team strengths against each other, a rating is calculated
	per team by dividing the team's	strength by the opposing team's strength
	(rating = ownStrength / opposingStrength).

	 Example (continued):
	 CT rating: 3.73/7.75 = 0.48
	 T rating: 7.75/3.73 = 2.08

	Just in case you want to know: the old ratings would have been 0.88 for
	the CT team and 1.125 for the T team. Just imagine to add player F (0:10)
	to the CT team and look what happens to the old and new ratings. I found
	the new ratings to represent the teams much better.

	But team rating is not all. Even the best team might fail. That's why PTB
	doesn't limit itself to those kill/death ratios. PTB also compares the number
	of team wins and notes the number of consecutive rounds one by a team as
	additional indicators for team efficiency.
	
	To decide, if there if there is a winning team situation, PTB then evaluates
	the following equations:

	o team rating >= PTB_MINRATING (stronger team) and score > PTB_MAXDIFF
	o team rating >= PTB_MINRATING (stronger team) and streak > PTB_MAXSTREAK
	o team rating >= PTB_MINRATING (stronger team) and more players
	o team rating >= PTB_MAXRATING (way stronger team)
	o team rating >= PTB_SUPERRATING (overwhelming team strength)

	If any TWO of these criteria hold for one team, it is considered to
	be the winning team and actions are taken as appropriate.

	Note: Teams with a rating below PTB_MINRATING are considered as of equal
	or very similar strength. If team strengths are very similar, streaks
	and	scores are ignored. As soon as a team gets noticibly stronger
	than it's opposing team, scores, streaks and team size are checked.
	
	PTB balances the teams in one of two ways. If the losing team is smaller,
	PTB transfers a player from the winning to the losing team. Else two players
	are exchanged (switched) with each other. Valid targets for switches and
	transfers are dead players and players, who didn't change teams already
	this round. These limitations origin from the fact, that it's not possible
	to switch a living player with the help of console commands without killing
	him. Additionally CS currently does allow only one team change per round.
	Since PTB 1.5 all possible transfers or switches are evaluated for the
	switch or tranfer, that balances the teams the most. If no switch or transfer
	would improve team balance, switching or transfering is skipped.

	Configuration
	-------------
	PTB Required Variable Settings:
	allow_client_exec 1 - needed to switch players' teams
	admin_balance_teams 0 - switch other team balancing methods off
	mp_autoteambalance 0 - switch off inbuild CS team balancing
	mp_limitteams 10 - enable any team size difference

	PTB Commands:

	// misc
	amx_ptb - show PTB statistics (ACCESS_ALL)
	amx_ptb status - show all current settings (ACCESS_ALL)
	amx_ptb list - list of ptb commands (ACCESS_ALL)
	amx_ptb help - same as "list" (ACCESS_ALL)
	amx_ptb on - enable all options
	amx_ptb off - disable all options
	amx_ptb save - save all settings to vault.ini
	amx_ptb load - load all settings from vault.ini

	// team selection control
	amx_ptb limitjoin on|off (default: on) - team join control
	amx_ptb limitafter <value> (default: 0) - rounds after which teams limiting begins
	amx_ptb limitmin <value> (default: 0) - minimum players for team limiting
	amx_ptb maxsize <value> (default: 10) - maximum team size per team
	amx_ptb autorounds <value> (default: 3) - initial rounds without free team choice
	amx_ptb maxdiff <value> (default: 2) - maximum accepted team size difference
	amx_ptb wtjauto <value> (default: 2) - WTJ tries for auto-join
	amx_ptb wtjkick <value> (default: 3) - WTJ tries for kick
	amx_ptb kick on|off (default: on) - WTJ kicking
	amx_ptb savewtj on|off (default: off) - wtj.log writing

	// team balancing actions
	amx_ptb switch on|off (default: on) - team switching and transfers
	amx_ptb switchafter <value> (default: 0) - rounds after which switching begins
	amx_ptb switchmin <value> (default: 3) - minimum players on map for switching
	amx_ptb switchfreq <value> (default: 1) - maximum team switching frequency
	amx_ptb playerfreq <value> (default: 3) - maximum player switching frequency
	amx_ptb forceswitch <value> (default: 3) - forced switch delay (switching alive, if neccessary)
	amx_ptb deadonly on|off (default: on) - switch dead only or alive, too

	// messages
	amx_ptb tellwtj on|off (default: on) - tell about wtj tries
	amx_ptb announce on|off (default: on) - announcements
	amx_ptb sayok on|off (default: on) - "no action required" message
	amx_ptb typesay on|off (default: on) - globally switch use of typesay

	// team strength limits
	amx_ptb maxstreak <value> (default: 2) - maximum accepted team win streak
	amx_ptb maxscore <value> (default: 2) - maximum accepted team score difference
	amx_ptb minrating <value> (default: 1.5) - stronger team, if rating >= minrating
	amx_ptb maxrating <value> (default: 2.0) - way stronger team, if rating >= maxrating
	amx_ptb superrating <value> (default: 3.0) - overwhelming team strength, if rating >= superrating
	 (don't play with these, if you don't fully understand, what they mean)

	Configuration Using "vault.ini"
	-------------------------------
	PTB options can be set by manipulating the "option_..." variables at
	the top of the code, or by setting variables in AdminMod's vault.ini
	file. If you want to do the latter, these are the default settings
	as they should appear in the file (each option should stand at the
	beginning of a new line). Look at the comments besides the corresponding
	"options_..." code to find out what the options do. Atlernatively you can
	type "amx_ptb status" in the console of a running CS server to get some
	info on the settings.
	
	// team selection control
	PTB_LIMITJOIN    on
	PTB_LIMITAFTER   0
	PTB_LIMITMIN     0
	PTB_MAXSIZE      10
	PTB_MAXDIFF      2
	PTB_AUTOROUNDS   3
	PTB_WTJAUTO      2
	PTB WTJKICK      3
	PTB_KICK         on
	PTB_SAVEWTJ      off

	// team balancing actions
	PTB_SWITCH       on
	PTB_SWITCHAFTER  0
	PTB_SWITCHMIN    3
	PTB_SWITCHFREQ   1
	PTB_PLAYERFREQ   3
	PTB_FORCESWITCH  3
	PTB_DEADONLY     on

	// messages
	PTB_TELLWTJ      on
	PTB_ANNOUNCE     on
	PTB_SAYOK        on
	PTB_TYPESAY      on

	// team strength limits
	PTB_MAXSTREAK    2
	PTB_MAXSCORE     2
	PTB_MINRATING    1.5
	PTB_MAXRATING    2.0
	PTB_SUPERRATING  3.0

	Latest Changes
	--------------
	Version 1.6
	+ NEW: added new winning team condition (PTB_SUPERRATING)
	+ NEW: new console commands "amx_ptb list" and "amx_ptb help"
	+ NEW: new console command "amx_ptb save" saves settings to "vault.ini"
	+ NEW: new console command "amx_ptb load" loads settings from "vault.ini"
	+ NEW: grouped settings and added missing console commands
	+ NEW: rounds before team limiting setting (PTB_LIMITAFTER, default: 0)
	+ NEW: minimum players for team limiting setting (PTB_LIMITMIN, default: 0)
	+ NEW: maximum team size setting (PTB_MAXSIZE, default: 10)
	+ NEW: rounds before switching setting (PTB_SWITCHAFTER, default: 0)
	+ NEW: minimum players for switching setting (PTB_SWITCHMIN, default: 3)
	+ NEW: maximum switch frequency setting (PTB_SWITCHFREQ, default: 1)
	+ NEW: tell about WTJ tries setting (PTB_TELLWTJ, default: on)
	+ NEW: maximum player switch frequency setting (PTB_PLAYERFREQ, default: 3)
	+ CHANGED: replaced all option_ identifiers with their PTB_ counterparts
	+ CHANGED: further code cleaning
	+ FIXED: hopefully improved the "negative result on fdiv" workaround

	Version 1.5
	+ NEW: new team rating calculation method with much improved strength indication
	+ NEW: much improved team switching code, choosing best balancing switch/transfer
	  (thanks to NtroP for providing part of the code for this)
	+ NEW: reading configuration options from vault.ini (see section about vault.ini)
	+ NEW: option to globally disable typesay messages (PTB_TYPESAY, default: 1)
	+ NEW: wtj tries for autojoin setting (PTB_WTJAUTO, default: 2)
	+ NEW: wtj tries for kick setting (PTB_WTJKICK, default: 3)
	+ NEW: max team size difference setting (PTB_MAXDIFF, default: 2)
	+ NEW: auto join only in the first x rounds setting (PTB_AUTOROUNDS, default: 3)
	+ NEW: force switch after x unsuccessful tries setting (PTB_FORCESWITCH, default: 3)
	+ CHANGED: changed file name to plugin_logd_cs_ptb.sma
	+ CHANGED: added new documentation topics and updated the old ones
	+ REMOVED: removed option_maxrandom, as there is no random switching anymore
	+ FIXED: didn't catch all win conditions (e. g. Bomb_Defused was not counted)

	Version 1.4a (never published intermediate version)
	+ FIXED: Some semicolons missed, causing problems with AdminMod 2.50.26 (Metamod 1.12 version)

	Version 1.4
	+ FIXED: PTB didn't work the first time a player chooses a team
	+ NEW: PTB sets team strength to team size, if team has no single kill yet
	  (fix for team inbalance in the first rounds of a map)
	+ NEW: switching/transfering alive, if dead only is not possible repeatedly

	Version 1.3
	+ NEW: announcing team switch/transfer failures as typesay message
	+ NEW: added "sayok" option
	+ FIXED: completely reworked the options interface, which didn't work at all
	+ WORKAROUND: retrying calculation for negative results on division of 2 positive numbers
	+ REMOVED: command amx_ptb_stats no longer exists

	Version 1.2a
	+ FIXED: compile error with Adminmod 2.50.26 (plugin_connect syntax change in Adminmod?)

	Version 1.2
	+ NEW: display WTJ count
	+ NEW: kick for WTJ count 10 or greater
	+ NEW: force players into losing/smaller team for WTJ count > 3
	+ NEW: save player name and WONID in wtj.log for WTJ count > 3
	+ NEW: changed rating to incorporate team sizes
	+ NEW: added switch dead only mode
	+ NEW: added lots of configuration variables
	+ NEW: added team advantage announcements
	+ FIXED: trying to switch player that had already changed teams in the same round

	KNOWN BUGS
	----------
	- sometimes dividing 2 positive fixed numbers yields a negative number,
	  the chosen workaround might not always do it
	- playercount() doesn't always give the correct number
	- playerinfo() doesn't always give correct team

	SUGGESTIONS & TODO
	------------------
	- keep spectators out of PTB_LIMITMIN and PTB_SWITCHMIN
	- configurable colors
	- configurable/translatable text messages
	- modified ying-yang logo
	- invent map type factors (de_, cs_, as_)
	(PTB 1.5 introduces new switching code, it remains to be seen,
	if this makes other potential changes obsolete)

*/

#pragma dynamic 32768

#include <amxmod>
#include <fixed>
#include <logd>


new STRING_VERSION[64] = "1.6";

#define ACCESS_BALANCE ADMIN_LEVEL_A
#define ACCESS_CONSOLE ADMIN_SLAY
#define ACCESS_ALL 0

#define TS 1
#define CTS 2
#define AUTO_TEAM 5

// defaults
new MIN_RATING[10] = "1.5";
new MAX_RATING[10] = "2.0";
new SUPER_RATING[10] = "3.0";

// options

// team selection control
new PTB_LIMITJOIN = 1; // set limits on team joining
new PTB_LIMITAFTER = 0; // number of rounds after which teams limiting begins
new PTB_LIMITMIN = 0; // number of minimum players on map for team limiting
new PTB_MAXSIZE = 10; // maximum team size per team
new PTB_MAXDIFF = 2; // maximum team size difference
new PTB_AUTOROUNDS = 3; // number of rounds into match, which allow autojoin only
new PTB_WTJAUTO = 2; // wtj tries needed to become autojoined
new PTB_WTJKICK = 3; // wtj tries needed to become kicked
new PTB_KICK = 1; // kick for wtj counts
new PTB_SAVEWTJ = 0; // save wtjs to wtj.txt

// team balancing actions
new PTB_SWITCH = 1; // switch/transfer players
new PTB_SWITCHAFTER = 0; // number of rounds after which switching begins
new PTB_SWITCHMIN = 3; // number of minimum players on map for switching
new PTB_SWITCHFREQ = 1; // relative next possible switch round
new PTB_PLAYERFREQ = 3; // relative next possible switch round for player
new PTB_FORCESWITCH = 3; // number of tries after which PTB switches alive, if neccessary
new PTB_DEADONLY = 1; // switch dead only

// messages
new PTB_TELLWTJ = 1; // tell about wtj tries
new PTB_ANNOUNCE = 1; // announce team status at beginning of round
new PTB_SAYOK = 1; // announce team status, if teams are alright
new PTB_TYPESAY = 1; // use typesay

// team strength limits
new PTB_MAXSTREAK = 2; // max allowed team win streak
new PTB_MAXSCORE = 2; // max allowed team score difference
new fixed:PTB_MINRATING; // minimum critical team rating
new fixed:PTB_MAXRATING; // maximum critical team rating
new fixed:PTB_SUPERRATING; // super critical team rating

new sortedTeams[3][32 + 1];

new validTargetCounts[3];
// sorted player indices are 0-based
new sortedValidTargets[3][32 + 1];

new kills[32 + 1] = { 0, ... };
new deaths[32 + 1] = { 0, ... };

new teamKills[3] = { 0, ... };
new teamDeaths[3] = { 0, ... };
new teamScores[3] = { 0, ... };
new winStreaks[3] = { 0, ... };

new wtConditions[3];
new winnerTeam = 0;
new loserTeam = 0;

new roundCounter = 0;
new betweenRounds = true;

new fixed:ctKD;
new fixed:tKD;
new fixed:ctStrength;
new fixed:tStrength;
new fixed:ctRating;
new fixed:tRating;

// player arrays are 1-based, there is no player 0
new isChoosingTeam[32 + 1];
new isBeingTransfered[32 + 1];
new playerTeam[32 + 1];
new lastRoundSwitched[32 + 1];
new wtjCount[32 + 1];
new teamCounts[3] = { 0, ... };

new lastSwitchRound = 0;
new couldNotSwitchCounter = 0;

new lastTeamBalanceCheck[64];

new vaultfile[64]

in_vault(var[]) {
	new i = 1
	new line[64 + 1]
	new lineLen
	new Cvar[32]
	new Cval[32]
	if(!(file_exists(vaultfile))) 
		return -1
	
	while( read_file(vaultfile,i,line,64,lineLen) ) {
		parse(line,Cvar,32,Cval,32)
		if(equal(Cvar,var)) 
			return i
		i++
		if(equal(Cvar,"//",2))
			continue
	}
	return -1
}

log(text[])
	log_message(text)

set_vaultdata(var[],val[]) {
	new linenum = in_vault(var)
	new line[64]
	format(line,64,"%s %s",var,val)
	write_file(vaultfile,line,linenum)
	return PLUGIN_HANDLED
}

get_vaultdata(var[],val[],len) {
	new linenum = in_vault(var)
	if (linenum == -1)
		return 0
	new line[64]
	new lineLen
	new FCvar[32]
	new Fval[32]
	read_file(vaultfile, linenum, line, 64, lineLen)
	parse(line,FCvar,32,Fval,32)
	copy(val,len,Fval)
	return 1
}

check_param(param[]) {
	if(equal(param,"1")) 
		return 1
	if(equal(param,"on"))
		return 1
	return 0
}

fixed:fabs(fixed:n) {
	if (n >= fixed(0)) return n;
	return -n;
}


fixed:fdivWorkaround(fixed:nom, fixed:denom) {
	// return nom, if denom == 0
	if (denom == fixed(0)) return nom;

	// handle negative result (no idea, if negativ results are reliable ...)
	if (nom >= fixed(0) && denom < fixed(0)) return fdiv(nom, denom);
	if (nom < fixed(0) && denom > fixed(0)) return fdiv(nom, denom);

	// workaround for negative results on dividing 2 positive or 2 negative numbers
	nom = fabs(nom);
	denom = fabs(denom);
	new fixed:result = nom;
	new MAX_TRIES = 20;
	new tries = 0;
	while (tries < MAX_TRIES) {
		result = fdiv(nom, denom);
		if (fmul(result, denom) == nom) break;
		if (result >= fixed(0)) break;
		++tries;
	}
	return result;
}

doTypesay(string[], duration, r, g, b) {
	if (!PTB_TYPESAY)
		return PLUGIN_HANDLED
	set_hudmessage(r, g, b, 0.05, 0.70, 0, float(duration), 6.0, 6.0, 0.15, 3) 
	show_hudmessage(0,"%s",string)
	return PLUGIN_HANDLED
}

transferPlayer(id) {
	//say("transferPlayer");
	
	new name[32];
	get_user_name(id, name, 32);

	isBeingTransfered[id] = 1;

	//say("Transferring Player:");
	//say(Name);

	if (playerTeam[id] == TS) {
		client_cmd(id, "chooseteam; menuselect 2; menuselect 5");
	}
	else {
		client_cmd(id, "chooseteam; menuselect 1; menuselect 5");
	}

	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
	client_cmd(id, "wait; wait; wait; slot6, wait; wait; wait; slot6");
}

actAtEndOfRound()
{
	//say("act");
	if (!PTB_SWITCH) return;

	// skip switching for the first few rounds
	if (roundCounter <= PTB_SWITCHAFTER) return;

	// honor switch frequency setting
	if (roundCounter - lastSwitchRound < PTB_SWITCHFREQ) return;

	// skip switching for a small number of players
	if (get_playersnum() < PTB_SWITCHMIN) return;

	client_print(0,print_chat,"PTB: Round ended, checking teams.");

	checkTeamBalance();

	if (winnerTeam != 0) {
		sortTeam(CTS);
		sortTeam(TS);
		
		if (teamCounts[winnerTeam] <= teamCounts[loserTeam]) doSwitch();
		else if (teamCounts[loserTeam] < teamCounts[winnerTeam]) doTransfer();
		
		//new text[64];
		//format(text, 64, "Winners: %d, Losers: %d", winners, losers);
		//say(text);
	}
}

createValidTargets(theTeam, deadonly) {
	new Name[32];
	new UserId;
	new dead;

	//say("createValidTargets");
	new n = 0;
	for (new i = 0; i < teamCounts[theTeam]; ++i) {
//		if (0 == playerinfo(sortedTeams[theTeam][i], Name, 32, WonId, UserId, team, dead)) {
//			//say("No playerinfo ...");
//		}
		get_user_name(sortedTeams[theTeam][i],Name,32)
		UserId = get_user_userid(sortedTeams[theTeam][i])
		if( is_user_alive(UserId) )
			dead = 0
		else 
			dead = 1


		//format(Name, 32, "Index: %d, sortedTeamIndex: %d, dead: %d", i , sortedTeams[theTeam][i], dead);
		//say(Name);
		if (dead == 0 && deadonly) {
			//say("player alive");
			continue;
		}
		if (
			 (roundCounter == lastRoundSwitched[sortedTeams[theTeam][i]])
			 || (roundCounter - lastRoundSwitched[sortedTeams[theTeam][i]] < PTB_PLAYERFREQ)
		) {
			//say("Player switched this round...")
			continue;
		}
		//say("player dead and not switched this round");
		sortedValidTargets[theTeam][n] = sortedTeams[theTeam][i];
		++n;
	}
	validTargetCounts[theTeam] = n;
}

sortTeam(theTeam) {
	//say("sortTeam");

	// create list of players
	new n = 0;
	for (new i = 1; i <= 32; ++i) {
		if (playerTeam[i] != theTeam) continue;
		sortedTeams[theTeam][n] = i;
		++n;
	}
	
	// do a selection sort
	new count = teamCounts[theTeam];
	for (new i = count - 1; i > 0; --i) {
		for (new k = i - 1; k >= 0; --k) {
			// compare players
			if (
				(kills[sortedTeams[theTeam][k]] < kills[sortedTeams[theTeam][i]])
				|| (
					(kills[sortedTeams[theTeam][k]] == kills[sortedTeams[theTeam][i]])
					&& (deaths[sortedTeams[theTeam][k]] > deaths[sortedTeams[theTeam][i]])
				)
			) {
				// swap
				new t;
				t = sortedTeams[theTeam][k];
				sortedTeams[theTeam][k] = sortedTeams[theTeam][i];
				sortedTeams[theTeam][i] = t;
			}
		}
	}
}


fixed:score(team, toBeAdded, toBeRemoved) {
	new fixed:sumKD = fixed(0);

	for (new player = 1; player <= 32; ++player) {
		if (
			(playerTeam[player] != team && player != toBeAdded)
			|| player == toBeRemoved
		) continue;
		sumKD += fdivWorkaround(fixed(kills[player]), fixed(deaths[player]));
	}

	new fixed:strength = fixed(teamCounts[team]);
	if (sumKD > fixed(0)) strength = fmul(strength, sumKD);

	return strength;
}

say(text[])
	client_print(0,print_chat,text)

doSwitch() {
	//say("doSwitch");
	new text[128];

	//displayStatistics(true);

	// don't switch, if at least one team is empty
	if (teamCounts[winnerTeam] == 0 || teamCounts[loserTeam] == 0) {
		text = "PTB: Can't switch players, need players in each team.";
		doTypesay(text, 5, 10, 255, 10);
		say(text);
		return;
	}

	// don't switch, if winner is alone (RULER!!!)
	if (teamCounts[winnerTeam] == 1) {
		text = "PTB: Won't switch players, best player makes the winning team.";
		doTypesay(text, 5, 10, 255, 10);
		say(text);
		return;
	}

	// don't switch, if both teams are full
	if (teamCounts[winnerTeam] == PTB_MAXSIZE && teamCounts[loserTeam] == PTB_MAXSIZE) {
		text = "PTB: Can't switch players, both teams are full.";
		doTypesay(text, 5, 10, 255, 10);
		say(text);
		return;
	}

	if (!PTB_DEADONLY || couldNotSwitchCounter > PTB_FORCESWITCH) {
		// choose from random top or bottom x
		createValidTargets(winnerTeam, 0);
		createValidTargets(loserTeam, 0);

		if (validTargetCounts[winnerTeam] == 0 || validTargetCounts[loserTeam] == 0) {
			copy(text, 64, "PTB: Can't switch players, need valid target in each team.");
			doTypesay(text, 5, 10, 255, 10);
			say(text);
			++couldNotSwitchCounter;
			return;
		}
	}
	else {
		//say("switch dead");
		createValidTargets(winnerTeam, 1);
		createValidTargets(loserTeam, 1);

		if (validTargetCounts[winnerTeam] == 0 || validTargetCounts[loserTeam] == 0) {
			++couldNotSwitchCounter;
			if (couldNotSwitchCounter > PTB_FORCESWITCH) {
				say("PTB: Couldn't switch dead, switching alive.");
				doSwitch();
				return;
			}
			copy(text, 64, "PTB: Can't switch players, need valid target in each team.");
			doTypesay(text, 5, 10, 255, 10);
			say(text);
			return;
		}
	}

	// Now search through the possible 1 to 1 swaps to equalize the strength as much as possible
	new fixed:closestScore = fabs(score(winnerTeam, -1, -1) - score(loserTeam, -1, -1));
	new winner = 0;
	new loser = 0;

	for (new w = 0; w < validTargetCounts[winnerTeam]; ++w) {
		new toLoser = sortedValidTargets[winnerTeam][w];
		for (new l = 0; l < validTargetCounts[loserTeam]; ++l) {
			new toWinner = sortedValidTargets[loserTeam][l];
			new fixed:myScore = fabs(score(winnerTeam, toWinner, toLoser) - score(loserTeam, toLoser, toWinner));
			if (myScore < closestScore) {
				closestScore = myScore;
				winner = toLoser;
				loser = toWinner;
			}
		}
	}

	if (winner == 0 && loser == 0) {
		copy(text, 64, "PTB: No switch would improve team balancing.");
		doTypesay(text, 5, 10, 255, 10);
		say(text);
		return;
	}

	couldNotSwitchCounter = 0;
	lastSwitchRound = roundCounter;

	new winnerName[32];
	new loserName[32];

	//new debug[64];
	//format(debug, 64, "Winner Index: %d", winnerIndex);
	//say(debug);
	//format(debug, 64, "Loser Index: %d", loserIndex);
	//say(debug);

	get_user_name(winner, winnerName, 32);
	get_user_name(loser, loserName, 32);

	// if one team is full, first move the the player from the full team ...
	if (teamCounts[winnerTeam] == PTB_MAXSIZE){
		transferPlayer(winner);
		transferPlayer(loser);
	}
	else {
		transferPlayer(loser);
		transferPlayer(winner);
	}
	format(text,128,"PTB: Switching %s with %s.",winnerName,loserName)

	doTypesay(text, 5, 10, 255, 10);
	say(text);
}

doTransfer() {
	//say("doTransfer");
	new text[128];

	if (teamCounts[winnerTeam] == 0) return;

	//displayStatistics(true);

	if (teamCounts[loserTeam] == PTB_MAXSIZE) {
		copy(text, 128, "PTB: Can't transfer player, losing team is full.");
		doTypesay(text, 5, 10, 255, 10);
		say(text);
		return;
	}

	if (!PTB_DEADONLY || couldNotSwitchCounter > PTB_FORCESWITCH) {
		createValidTargets(winnerTeam, 0);

		if (validTargetCounts[winnerTeam] == 0) {
			copy(text, 128,"PTB: Can't transfer player, no valid target in winning team.");
			doTypesay(text, 5, 10, 255, 10);
			say(text);
			++couldNotSwitchCounter;
			return;
		}
	}
	else {
		//say("switch dead");
		createValidTargets(winnerTeam, 1);
		if (validTargetCounts[winnerTeam] == 0) {
			++couldNotSwitchCounter;
			if (couldNotSwitchCounter > PTB_FORCESWITCH) {
				say("PTB: Couldn't transfer dead, transferring alive.");
				doTransfer();
				return;
			}
			copy(text, 128, "PTB: Can't transfer player, no valid target in winning team.");
			doTypesay(text, 5, 10, 255, 10);
			say(text);
			return;
		}
	}

	new fixed:closestScore = fabs(score(winnerTeam, -1, -1) - score(loserTeam, -1, -1));
	new winner = -1;

	for (new w = 0; w < validTargetCounts[winnerTeam]; ++w) {
		new toLoser = sortedValidTargets[winnerTeam][w];
		new fixed:myScore = fabs(score(winnerTeam, -1, toLoser) - score(loserTeam, toLoser, -1));
		if (myScore < closestScore) {
			closestScore = myScore;
			winner = toLoser;
		}
	}

	if (winner == -1) {
		copy(text, 128, "PTB: No transfer would improve team balancing.");
		doTypesay(text, 5, 10, 255, 10);
		say(text);
		return;
	}

	couldNotSwitchCounter = 0;

	new winnerName[32];

	//new debug[64];
	//format(debug, 64, "Winner Index: %d", winnerIndex);
	//say(debug);

	get_user_name(winner, winnerName, 32);
	
	transferPlayer(winner);
	new TeamName[6+1]
	if (winnerTeam == CTS) copy(TeamName, 6, "Ts");
	else if (winnerTeam == TS) copy(TeamName, 6,"CTs");
	format(text, 128, "PTB: Transfering %s to the %s.",winnerName,TeamName)
	doTypesay(text, 5, 10, 255, 10);
	say(text);
}

checkTeamBalance() {
	//say("Checking team balance ...");

	get_time("%H:%M:%S",lastTeamBalanceCheck, 64);

	calcTeamScores();

	//ctStrength = fixed(teamCounts[CTS]);
	// workaround for match start team balancing
	// this is making only team size count,
	// as long as there is no single kill
	//if (ctKD != fixed(0)) ctStrength = fmul(ctKD, ctStrength);

	// replacing old strength calculation
	ctStrength = score(CTS, -1, -1);

	//tStrength = fixed(teamCounts[TS]);
	// workaround for match start team balancing
	// this is making only team size count,
	// as long as there is no single kill
	//if (tKD != fixed(0)) tStrength = fmul(tKD, tStrength);

	// replacing old strength calculation
	tStrength = score(TS, -1, -1);

	ctRating = fdivWorkaround(ctStrength, tStrength);
	tRating = fdivWorkaround(tStrength, ctStrength);
	
	wtConditions[TS] = 0;
	wtConditions[CTS] = 0;

	//new buffer[10];

	// compare scores for unequal rating scores
	if (teamScores[TS] - teamScores[CTS] > PTB_MAXSCORE && tRating >= PTB_MINRATING) {
		wtConditions[TS] = wtConditions[TS] + 1;
	}
	if (teamScores[CTS] - teamScores[TS] > PTB_MAXSCORE && ctRating >= PTB_MINRATING) {
		wtConditions[CTS] = wtConditions[CTS] + 1;
	}

	//say("// check streaks for unequal rating scores");
	// check streaks for unequal rating scores
	if (winStreaks[TS] > PTB_MAXSTREAK && tRating >= PTB_MINRATING) {
		wtConditions[TS] = wtConditions[TS] + 1;
	}
	if (winStreaks[CTS] > PTB_MAXSTREAK && ctRating >= PTB_MINRATING) {
		wtConditions[CTS] = wtConditions[CTS] + 1;
	}

	//say("// check ratings");
	// check ratings
	if (tRating >= PTB_MAXRATING) {
		wtConditions[TS] = wtConditions[TS] + 1;
	}
	if (ctRating >= PTB_MAXRATING) {
		wtConditions[CTS] = wtConditions[CTS] + 1;
	}
	
	//say("// check super ratings");
	// check ratings
	if (tRating >= PTB_SUPERRATING) {
		wtConditions[TS] = wtConditions[TS] + 1;
	}
	if (ctRating >= PTB_SUPERRATING) {
		wtConditions[CTS] = wtConditions[CTS] + 1;
	}
	
	//say("// check team sizes for unequal ratings");
	// check team sizes for unequal ratings
	if (teamCounts[TS] > teamCounts[CTS] && tRating >= PTB_MINRATING) {
		wtConditions[TS] = wtConditions[TS] + 1;
	}
	if (teamCounts[CTS] > teamCounts[TS] && ctRating >= PTB_MINRATING) {
		wtConditions[CTS] = wtConditions[CTS] + 1; 
	}
	
	//say("// check conditions");
	// check conditions
	if (wtConditions[TS] >= 2) {
		//say("TERRORISTS are the winning team ...");
		winnerTeam = TS;
		loserTeam = CTS;
	}
	else if (wtConditions[CTS] >= 2) {
		//say("COUNTER-TERRORISTS are the winning team ...");
		winnerTeam = CTS;
		loserTeam = TS;
	}
	else {
		winnerTeam = 0;
		loserTeam = 0;
		//if (wtConditions[CTS] > 0) say("COUNTER-TERRORIST team advantage ...");
		//else if (wtConditions[TS] > 0) say("TERRORIST team advantage ...");
		//else say("No winning team, no team advantage ...");
	}
}

public ChoosingTeam(id) {
	//say("ChoosingTeam");
	
	isChoosingTeam[id] = 1;
	
	return PLUGIN_CONTINUE;
}

public NotChoosingTeam(id) {
	//say("NotChoosingTeam");

	isChoosingTeam[id] = 0;
	
	return PLUGIN_CONTINUE;
}

manageWtjFile(UserIndex) {
	if (!PTB_SAVEWTJ) return;
	
	//log("Trying to write wtj.log ....");

	if (wtjCount[UserIndex] != 4) return;

	//log("wtj.log should be written to now ....");

	//new debug[64];
	//format(debug, 64, "PTB manageWtjFile");
	//log(debug);
	
	new Time[64];
	get_time("%H:%M:%S",Time, 64);

	new Map[64];
	get_mapname(Map, 64);

	new Name[32];
	new WonId;
	get_user_name(UserIndex,Name,32)
	WonId = get_user_wonid(UserIndex)
	new text[64];
	format(text, 64, "%s %s <%d> %s", Time, Name, WonId, Map);

	write_file("addons/amx/wtj.log", text);
}

public MenuSelect(id) {
	//say("MenuSelect");

	/*
	new s[64];
	new d[64];
	new n[32];
	get_user_name(id, n, 32)
	read_args(d, 64)
	format(s, 64, "%s (%d) selected menu %s (is choosing team: %d)", n, id, d, isChoosingTeam[id]);
	say(s);
	*/
	if (!isChoosingTeam[id]) return PLUGIN_CONTINUE;
	
	isChoosingTeam[id] = 0;

	//if (UserIndex < 1) return PLUGIN_CONTINUE;
	if (!PTB_LIMITJOIN) return PLUGIN_CONTINUE;

	// skip limiting for a few rounds into the map
	if (roundCounter <= PTB_LIMITAFTER) return PLUGIN_CONTINUE;

	// skip limiting for a small number of players
	if (get_playersnum() < PTB_LIMITMIN) return PLUGIN_CONTINUE;

	if (isBeingTransfered[id]) {
		//say("TRANSFER");
		isBeingTransfered[id] = 0;
		return PLUGIN_CONTINUE;
	}
	else {
		//say("NO TRANSFER");
	}

	new Data[64];
	new iNewTeam;

	read_argv(1,Data,64)
	iNewTeam = str_to_num(Data);

	// disallow free team choices in the first rounds of a map
	if (roundCounter <= PTB_AUTOROUNDS) iNewTeam = 5;
	
	new iOldTeam = playerTeam[id];
	
	new Name[32];
	get_user_name(id,Name,32)


	// prevent unwanted rejoining of the same team ...
	if (iNewTeam == playerTeam[id]) {
		//say("Preventing rejoining of the same team.");
		return PLUGIN_HANDLED;
	}

	checkTeamBalance();
	
	//displayStatistics(true);

	// debug
	//format(s, 64, "iOldTeam = %d, iNewTeam = %d, winnerTeam = %d, loserTeam = %d, CTS = %d, TS = %d, AUTO_TEAM = %d, #CTs = %d, #Ts = %d", iOldTeam, iNewTeam, winnerTeam, loserTeam, CTS, TS, AUTO_TEAM, teamCounts[CTS], teamCounts[TS]);
	//say(s);
	
	// check if player is trying to change teams
	if ((iNewTeam == CTS || iNewTeam == TS) && (iOldTeam == CTS || iOldTeam == TS)) {
		//say("Changing team directly.");
		new opposingTeam = iNewTeam == CTS ? TS : CTS;
		if (
			teamCounts[iNewTeam] != 0
			&& teamCounts[opposingTeam] < PTB_MAXSIZE
			&& (
				(winnerTeam != 0 && iNewTeam == winnerTeam)
				|| (teamCounts[iNewTeam] >= teamCounts[opposingTeam])
			)
		) {
			// player is wtjing
			new text[64];
			wtjCount[id] += 1;
			if (wtjCount[id] >= PTB_WTJKICK && PTB_KICK) {
				format(text, 80, "PTB: Kicking %s for a WTJ count of %d).", Name, wtjCount[id]);
				doTypesay(text, 5, 10, 255, 10);
				say(text);
				new uid = get_user_userid(id)
				server_cmd("kick #%d",uid)
				return PLUGIN_HANDLED;
			}
			if (PTB_TELLWTJ) {
				if (iNewTeam == CTS) {
					format(text, 80, "PTB: The CTs are strong enough, %s (WTJ count: %d).", Name, wtjCount[id]);
					doTypesay(text, 5, 10, 10, 255);
				}
				else {
					format(text, 80, "PTB: The Ts are strong enough, %s (WTJ count: %d).", Name, wtjCount[id]);
					doTypesay(text, 5, 255, 10, 10);
				}
				say(text);
			}
			return PLUGIN_HANDLED;
		}
		
		// check for maximum team size
		if (teamCounts[iNewTeam] + 1 > PTB_MAXSIZE) {
			client_print(id,print_chat,"PTB: Maximum team size prohibits team change.");
			return PLUGIN_HANDLED;
		}

		// check team size difference limits
		if ((teamCounts[iNewTeam] + 1) - (teamCounts[opposingTeam] - 1) > PTB_MAXDIFF) {
			client_print(id,print_chat,"PTB: Maximum team size difference prohibits team change.");
			return PLUGIN_HANDLED;
		}
	}
	else if (iNewTeam == CTS || iNewTeam == TS) {
		//say("Joining team directly.");
		new opposingTeam = iNewTeam == CTS ? TS : CTS;
		if (
			teamCounts[iNewTeam] != 0
			&& teamCounts[opposingTeam] < PTB_MAXSIZE
			&& (
				(winnerTeam != 0 && iNewTeam == winnerTeam) // && teamCounts[winnerTeam] >= teamCounts[opposingTeam])
				|| (winnerTeam == 0 && teamCounts[iNewTeam] > teamCounts[opposingTeam])
			)
		) {
			new text[64];
			wtjCount[id] += 1;
			if (wtjCount[id] >= PTB_WTJKICK && PTB_KICK) {
				format(text, 80, "PTB: Kicking %s for a WTJ count of %d).", Name, wtjCount[id]);
				doTypesay(text, 5, 10, 255, 10);
				say(text);
				new uid = get_user_userid(id)
				server_cmd("kick #%d",uid)
				return PLUGIN_HANDLED;
			}
			if (iNewTeam == CTS) {
				if (wtjCount[id] > PTB_WTJAUTO) {
					manageWtjFile(id);
					format(text, 80, "PTB: Forcing %s to the Ts (WTJ count: %d).", Name, wtjCount[id]);
					client_cmd(id, "menuselect 1");
					doTypesay(text, 5, 10, 255, 10);
					say(text);
				}
				else {
					if (PTB_TELLWTJ) {
						format(text, 80, "PTB: The CTs are strong enough, %s (WTJ count: %d).", Name, wtjCount[id]);
						doTypesay(text, 5, 10, 10, 255);
						say(text);
					}
				}
			}
			else {
				if (wtjCount[id] >= PTB_WTJAUTO) {
					manageWtjFile(id);
					format(text, 80, "PTB: Forcing %s to the CTs (WTJ count: %d).", Name, wtjCount[id]);
					client_cmd(id, "menuselect 2");
					doTypesay(text, 5, 10, 255, 10);
					say(text);
				}
				else {
					if (PTB_TELLWTJ) {
						format(text, 80, "PTB: The Ts are strong enough, %s (WTJ count: %d).", Name, wtjCount[id]);
						doTypesay(text, 5, 255, 10, 10);
						say(text);
					}
				}
			}
			return PLUGIN_HANDLED;
		}

		// check for maximum team size
		if (teamCounts[iNewTeam] + 1 > PTB_MAXSIZE) {
			client_print(id,print_chat,"PTB: Maximum team size prohibits team join.");
			return PLUGIN_HANDLED;
		}

		// check team size difference limits
		if ((teamCounts[iNewTeam] + 1) - (teamCounts[opposingTeam]) > PTB_MAXDIFF) {
			client_print(id,print_chat,"PTB: Maximum team size difference prohibits team join.");
			return PLUGIN_HANDLED;
		}
	}
	else if (iNewTeam == AUTO_TEAM && (iOldTeam == CTS || iOldTeam == TS)) {
		//say("Changing team automatically.");

		new opposingTeam = iOldTeam == CTS ? TS : CTS;
		if (
			teamCounts[opposingTeam] != 0
			&& (
				opposingTeam == PTB_MAXSIZE
				|| (loserTeam != 0 && iOldTeam == loserTeam)
				|| (loserTeam == 0 && teamCounts[iOldTeam] <= teamCounts[opposingTeam])
				|| ((teamCounts[opposingTeam] + 1) - (teamCounts[iOldTeam] - 1) > PTB_MAXDIFF)
			)
		) {
			//say("Aborting automatic team change.");
			// debug
			//new text[64];
			//format(text, 64, "Aborting auto-joining team %d (winnerTeam = %d) teamCounts[iOldTeam] %d teamCounts[opposingTeam] %d", opposingTeam, winnerTeam, teamCounts[iOldTeam], teamCounts[opposingTeam]);
			//say(text);

			// close menu
			return PLUGIN_HANDLED;
		}

		iNewTeam = opposingTeam;
		
		// debug
		//new text[64];
		//format(text, 64, "Auto-joining team %d. (winnerTeam = %d)", iNewTeam, winnerTeam);
		//say(text);

		// check for maximum team size
		if (teamCounts[iNewTeam] + 1 > PTB_MAXSIZE) {
			client_print(id,print_chat,"PTB: Maximum team size prohibits team change.");
			return PLUGIN_HANDLED;
		}

		if (iNewTeam == CTS) client_cmd(id, "menuselect 2");
		else client_cmd(id, "menuselect 1");

		return PLUGIN_HANDLED;
	}
	else if (iNewTeam == AUTO_TEAM) {
		//say("Joining team automatically.");

		/* this is the "always smaller team" version
		if (teamCounts[CTS] < teamCounts[TS] || teamCounts[TS] == PTB_MAXSIZE) iNewTeam = CTS;
		else if (teamCounts[TS] < teamCounts[CTS] || teamCounts[CTS] == PTB_MAXSIZE) iNewTeam = TS;
		// both teams have same size ...
		else if (winnerTeam != 0 && teamCounts[loserTeam] < PTB_MAXSIZE) iNewTeam = loserTeam;
		else if (teamCounts[TS] == PTB_MAXSIZE) iNewTeam = CTS;
		else if (teamCounts[CTS] == PTB_MAXSIZE) iNewTeam = TS;
		else iNewTeam = random(2) + 1;
		*/

		// this version prefers the losing team (but still honors PTB_MAXDIFF)
		if (teamCounts[CTS] == PTB_MAXSIZE) iNewTeam = TS;
		else if (teamCounts[TS] == PTB_MAXSIZE) iNewTeam = CTS;
		else if ((teamCounts[CTS] + 1) - (teamCounts[TS]) > PTB_MAXDIFF) iNewTeam = TS;
		else if ((teamCounts[TS] + 1) - (teamCounts[CTS]) > PTB_MAXDIFF) iNewTeam = CTS;
		else if (winnerTeam != 0) iNewTeam = loserTeam;
		else if (teamCounts[CTS] < teamCounts[TS]) iNewTeam = CTS;
		else if (teamCounts[TS] < teamCounts[CTS]) iNewTeam = TS;
		// both teams have same size ...
		else iNewTeam = random(2) + 1;
		
		// debug
		//new text[64];
		//format(text, 64, "Auto-joining team %d. (winnerTeam = %d)", iNewTeam, winnerTeam);
		//say(text);
		
		// check for maximum team size
		if (teamCounts[iNewTeam] + 1 > PTB_MAXSIZE) {
			client_print(id,print_chat,"PTB: Maximum team size prohibits team join.");
			return PLUGIN_HANDLED;
		}

		if (iNewTeam == CTS) client_cmd(id, "menuselect 2");
		else client_cmd(id, "menuselect 1");

		return PLUGIN_HANDLED;
	}

	//say("Allow team join ...");

	return PLUGIN_CONTINUE;
}

public logd_team_action(team[],action[],optional[])
{
	//say("ptb_teamaction");

	if(
		containi(action, "CTs_Win") 
		|| containi(action, "Bomb_Defused") 
		|| containi(action, "Terrorists_Win") 
		|| containi(action, "Target_Bombed")) 
	{
//	new scores[64];
	new ctScore[64];
	new tScore[64];
//	new skip[64];
	new scoreStr[64]
	new c
	new t

//	client_print(0,print_chat,"%s: %s: %s",team,action,optional)

//	parse(action, skip, 64, scores, 64);
	parse(optional, ctScore, 64, tScore, 64);
//	client_print(0,print_chat,"%s: %s: %s",action,skip,scores)
	

	format(scoreStr,64,"")
//	c = contain(ctScore, "CT#") + 3;
//	t = contain(action, " T#") - 1
	c = 3
	t = strlen(ctScore)

//	client_print(0,print_chat,"%s: %i:%c %i:%c",ctScore,c,ctScore[c],t,ctScore[t])

	for(new i = 0; i<((t-c)+1);i++)
		format(scoreStr,64,"%s%c",scoreStr,ctScore[c+i])
	teamScores[CTS] = strtonum(scoreStr);

//	c = contain(action, " T#") + 3;
//	t = strlen(action)
	c = 2
	t = strlen(tScore)
	
	format(scoreStr,64,"")	
	for(new i = 0; i<((t-c)+1);i++)
		format(scoreStr,64,"%s%c",scoreStr,tScore[c+i])
	teamScores[TS] = strtonum(scoreStr);
//	client_print(0,print_chat,"%s: %i:%c %i:%c",tScore,c,tScore[c],t,tScore[t])
	// count won rounds in a row
/*
		if (strmatch(action, "CTs_Win", strlen("CTs_Win"))
			|| strmatch(action, "Bomb_Defused", strlen("Bomb_Defused"))
		) {
*/
	if (!betweenRounds) {
		// prevent "bomb explodes after ts win" situations
		// from producing 2 wins in the same round
		if (equal(team, "CT", strlen("CT"))) {
			if (winStreaks[CTS] < 0) {
				// reset counters
				winStreaks[CTS] = 1;
				winStreaks[TS] = -1;
			}
			else {
				winStreaks[CTS] += 1;
				winStreaks[TS] -= 1;
			}
		}
		else if (equal(team, "TERRORIST", strlen("TERRORIST"))) {
			if (winStreaks[TS] < 0) {
				// reset counters
				winStreaks[CTS] = -1;
				winStreaks[TS] = 1;
			}
			else {
				winStreaks[CTS] -= 1;
				winStreaks[TS] += 1;
			}
		}

		actAtEndOfRound();
	}

	}

	// clear flags
	betweenRounds = true;


	return PLUGIN_CONTINUE;
}


public  logd_world_action(action[]) {


	if (equal(action, "Round_Start")) {
		++roundCounter;
		betweenRounds = false;
		//set_timer("announceStatus", 15, 1);
		announceStatus();
	}
	
	return PLUGIN_CONTINUE;
}

public logd_team_select(id,team[]) {
	//say("ptb_teamselection");

	//displayStatistics(true);
	new teamI;

	

	//new debug[64];
	//format(debug, 64, "PTB: ptb_teamselection: %s", data);
	//log(debug);
	
	if (equal(team, "TERRORIST")) teamI = TS;
	else if (equal(team, "CT")) teamI = CTS;
	else teamI = 0;
	
	//if (team == 0) say("PTB: ptb_teamselection: Team is 0.");
	

	if (playerTeam[id] == TS || playerTeam[id] == CTS) teamCounts[playerTeam[id]] -= 1;
	if (teamI == TS || teamI == CTS) teamCounts[teamI] += 1;

	playerTeam[id] = teamI;

	lastRoundSwitched[id] = roundCounter;
	
	//say("ptb_teamselection");
	//say(data);

	//displayStatistics(true);

	return PLUGIN_CONTINUE;
}

public logd_kill(id1,id2,weapon[],optional[])
{
	//say("ptb_kill");
	if (!is_user_connected(id1)) {
		return PLUGIN_CONTINUE;
	}

	kills[id1] = kills[id1] + 1;
	deaths[id2] = deaths[id2] + 1;

	//if (playerTeam[iWinner] != CTS && playerTeam[iWinner] != TS) log("PTB: Kill without team!");
	//if (playerTeam[iLoser] != CTS && playerTeam[iLoser] != TS) log("PTB: Death without team!");

	return PLUGIN_CONTINUE;
}

calcTeamScores()
{
	//say("calcTeamScores");

	teamDeaths[CTS] = 0;
	teamDeaths[TS] = 0;
	teamKills[CTS] = 0;
	teamKills[TS] = 0;
	
	for (new i = 1; i <= 32; ++i) {
		new team = playerTeam[i];
		if (team == CTS || team == TS) {
			teamKills[team] += kills[i];
			teamDeaths[team] += deaths[i];
		}
	}
	
	ctKD = fdivWorkaround(fixed(teamKills[CTS]), fixed(teamDeaths[CTS]));
	tKD = fdivWorkaround(fixed(teamKills[TS]), fixed(teamDeaths[TS]));

	//say("calcTeamScores DONE");
}

announceStatus() {

	if (!PTB_ANNOUNCE) return;
	    
	checkTeamBalance();
	
	new text[64];
	
	if (winnerTeam == TS) {
		format(text, 64, "PTB: The COUNTER-TERRORIST team could use some support.");
		doTypesay(text, 5, 10, 10, 255);
		say("PTB: The COUNTER-TERRORIST team could use some support.");
	}
	else if (winnerTeam == CTS) {
		format(text, 64, "PTB: The TERRORIST team could use some support.");
		doTypesay(text, 5, 255, 10, 10);
		say("PTB: The TERRORIST team could use some support.");
	}
	else if (wtConditions[TS] > wtConditions[CTS]) {
		format(text, 64, "PTB: Observing TERRORIST team advantage.");
		doTypesay(text, 5, 10, 10, 255);
		say("PTB: Observing TERRORIST team advantage.");
	}
	else if (wtConditions[CTS] > wtConditions[TS]) {
		format(text, 64, "PTB: Observing COUNTER-TERRORIST team advantage.");
		doTypesay(text, 5, 255, 10, 10);
		say("PTB: Observing COUNTER-TERRORIST team advantage.");
	}
	else if (PTB_SAYOK) {
		format(text, 64, "PTB: Teams look fine, no action required.");
		doTypesay(text, 5, 255, 255, 255);
		say("PTB: Teams look fine, no action required.");
	}

	//displayStatistics();
}

public amx_ptb(id) {
	new Data[64];
	new Command[64];
	new Argument[64];
	new s[64];
	new ls[128]

	read_args(Data, 32)
	copy(Command,64,"");
	copy(Argument,64,"");

	format(s, 64, "PTB: Ptahhotep's Team Balancer %s", STRING_VERSION);

	if (!strlen(Data)) {
		checkTeamBalance();
		displayStatistics(false,id);
		return PLUGIN_HANDLED;
	}
	else if (equal(Data, "on") || equal(Data, "1")) {
		if (!(get_user_flags(id)&ACCESS_BALANCE)) {
			client_print(id,print_console,"PTB: You are not allowed to use this command.");
			return PLUGIN_HANDLED;
		}
		PTB_LIMITJOIN = 1;
		PTB_SWITCH = 1;
		PTB_ANNOUNCE = 1;
		client_print(id,print_console,"PTB: Enabled all PTB actions.");
		return PLUGIN_HANDLED;
	}
	else if (equal(Data, "off") || equal(Data, "0")) {
		if (!(get_user_flags(id)&ACCESS_BALANCE)) {
			client_print(id,print_console,"PTB: You are not allowed to use this command.");
			return PLUGIN_HANDLED;
		}
		PTB_SWITCH = 0;
		PTB_ANNOUNCE = 0;
		PTB_LIMITJOIN = 0;
		client_print(id,print_console,"PTB: Disabled all PTB actions.");
		return PLUGIN_HANDLED;
	}
	else if (equal(Data, "save")) {
		if (!(get_user_flags(id)&ACCESS_BALANCE)) {
			client_print(id,print_console,"PTB: You are not allowed to use this command.");
			return PLUGIN_HANDLED;
		}
		saveSettings();
		client_print(id,print_console,"PTB: Saved all PTB settings to ^"%s^".",vaultfile);
		return PLUGIN_HANDLED;
	}
	else if (equal(Data, "load")) {
		if (!(get_user_flags(id)&ACCESS_BALANCE)) {
			client_print(id,print_console,"PTB: You are not allowed to use this command.");
			return PLUGIN_HANDLED;
		}
		loadSettings();
		client_print(id,print_console,"PTB: Loaded all PTB settings from ^"%s^".",vaultfile);
		return PLUGIN_HANDLED;
	}
	else if (equal(Data, "list") || equal(Data, "help")) {
		client_print(id,print_console,"PTB: Available Commands:");
		client_print(id,print_console,"PTB: Team Join Control: ^"limitjoin^", ^"limitafter^", ^"limitmin^", ^"maxsize^", ^"autorounds^",");
		client_print(id,print_console,"PTB: ^"maxdiff^", ^"wtjauto^", ^"wtjkick^", ^"kick^", ^"savewtj^"");
		client_print(id,print_console,"PTB: Team Balancing Actions: ^"switch^", ^"switchafter^", ^"switchmin^", ^"switchfreq^", ^"playerfreq^",");
		client_print(id,print_console,"PTB: ^"forceswitch^", ^"deadonly^"");
		client_print(id,print_console,"PTB: Team Strength Limits: ^"maxstreak^", ^"maxscore^", ^"minrating^", ^"maxrating^", ^"superrating^"");
		client_print(id,print_console,"PTB: Messages: ^"tellwtj^", ^"announce^", ^"sayok^", ^"typesay^"");
		client_print(id,print_console,"PTB: Misc: ^"^", ^"status^", ^"list^", ^"help^", ^"on^", ^"off^", ^"save^", ^"load^"");
		client_print(id,print_console,"PTB: To view all PTB settings, type ^"amx_ptb status^".");
		client_print(id,print_console,"PTB: To view or change a single PTB setting, type ^"amx_ptb <setting> <^"on^"|^"off^"|value>^".");
		client_print(id,print_console,"PTB: For PTB statistics, simply type ^"amx_ptb^".");
		return PLUGIN_HANDLED;
	}
	else {
		parse(Data, Command, 64, Argument, 64);
		if (strlen(Argument) && !(get_user_flags(id)&ACCESS_BALANCE)) {
			client_print(id,print_console,"PTB: You are not allowed to use this command.");
			return PLUGIN_HANDLED;
		}
	}


	// team selection control
	if (equal(Command, "status")) client_print(id,print_console,"PTB: ---------- Team Selection Control ----------");
	
	// PTB_LIMITJOIN
	if (equal(Command, "limitjoin") && strlen(Argument)) PTB_LIMITJOIN = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "limitjoin")) { if (PTB_LIMITJOIN) client_print(id,print_console,"PTB: (limitjoin) WTJ prevention is ON."); else client_print(id,print_chat,"PTB: (limitjoin) WTJ prevention is OFF."); }
	
	// PTB_LIMITAFTER
	if (equal(Command, "limitafter") && strlen(Argument)) {
		PTB_MAXDIFF = strtonum(Argument);
		if (PTB_LIMITAFTER < 0) PTB_LIMITAFTER = 0;
	}
	if (equal(Command, "status") || equal(Command, "limitafter")) {
		format(s, 64, "PTB: (limitafter) Team limiting starts after %d round(s).", PTB_LIMITAFTER);
		client_print(id,print_console,s);
	}

	// PTB_LIMITMIN
	if (equal(Command, "limitmin") && strlen(Argument)) {
		PTB_LIMITMIN = strtonum(Argument);
		if (PTB_LIMITMIN < 0) PTB_LIMITMIN = 0;
	}
	if (equal(Command, "status") || equal(Command, "limitmin")) {
		format(s, 64, "PTB: (limitmin) Team limiting needs at least %d player(s).", PTB_LIMITMIN);
		client_print(id,print_console,s);
	}

	// PTB_MAXSIZE
	if (equal(Command, "maxsize") && strlen(Argument)) {
		PTB_MAXSIZE = strtonum(Argument);
		if (PTB_MAXSIZE < 0) PTB_MAXSIZE = 0;
	}
	if (equal(Command, "status") || equal(Command, "maxsize")) {
		format(s, 64, "PTB: (maxsize) Maximum team size is %d player(s).", PTB_MAXSIZE);
		client_print(id,print_console,s);
	}

	// PTB_MAXDIFF
	if (equal(Command, "maxdiff") && strlen(Argument)) {
		PTB_MAXDIFF = strtonum(Argument);
		if (PTB_MAXDIFF < 1) PTB_MAXDIFF = 1;
	}
	if (equal(Command, "status") || equal(Command, "maxdiff")) {
		format(s, 64, "PTB: (maxdiff) Maximum team size difference is %d.", PTB_MAXDIFF);
		client_print(id,print_console,s);
	}

	// PTB_AUTOROUNDS
	if (equal(Command, "autorounds") && strlen(Argument)) {
		PTB_AUTOROUNDS = strtonum(Argument);
		if (PTB_AUTOROUNDS < 0) PTB_AUTOROUNDS = 0;
	}
	if (equal(Command, "status") || equal(Command, "autorounds")) {
		format(s, 64, "PTB: (autorounds) First %d rounds no free team choice.", PTB_AUTOROUNDS);
		client_print(id,print_console,s);
	}

	// PTB_WTJAUTO
	if (equal(Command, "wtjauto") && strlen(Argument)) {
		PTB_WTJAUTO = strtonum(Argument);
		if (PTB_WTJAUTO < 0) PTB_WTJAUTO = 0;
	}
	if (equal(Command, "status") || equal(Command, "wtjauto")) {
		format(s, 64, "PTB: (wtjauto) Auto-joining WTJ after %d tr(y/ies).", PTB_WTJAUTO);
		client_print(id,print_console,s);
	}

	// PTB_WTJKICK
	if (equal(Command, "wtjkick") && strlen(Argument)) {
		PTB_WTJKICK = strtonum(Argument);
		if (PTB_WTJKICK < 1) PTB_WTJKICK = 1;
	}
	if (equal(Command, "status") || equal(Command, "wtjkick")) {
		format(s, 64, "PTB: (wtjauto) Auto-kicking WTJ after %d tr(y/ies).", PTB_WTJKICK);
		client_print(id,print_console,s);
	}

	// PTB_KICK
	if (equal(Command, "kick") && strlen(Argument)) PTB_KICK = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "kick")) { if (PTB_KICK) client_print(id,print_console,"PTB: (kick) WTJ kicking is ON."); else client_print(id,print_console,"PTB: (kick) Kicking is OFF."); }
	
	// PTB_SAVEWTJ
	if (equal(Command, "savewtj") && strlen(Argument)) PTB_SAVEWTJ = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "savewtj")) { if (PTB_SAVEWTJ) client_print(id,print_console,"PTB: (savewtj) Saving to wtj.txt is ON."); else client_print(id,print_console,"PTB: (savewtj) Saving to wtj.txt is OFF."); }
	

	// team balancing actions
	if (equal(Command, "status")) client_print(id,print_console,"PTB: ---------- Team Balancing Actions ----------");

	// PTB_SWITCH
	if (equal(Command, "switch") && strlen(Argument)) PTB_SWITCH = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "switch")) { if (PTB_SWITCH) client_print(id,print_console,"PTB: (switch) Team switching is ON."); else client_print(id,print_console,"PTB: (switch) Team switching is OFF."); }
	
	// PTB_SWITCHAFTER
	if (equal(Command, "switchafter") && strlen(Argument)) {
		PTB_SWITCHAFTER = strtonum(Argument);
		if (PTB_SWITCHAFTER < 0) PTB_SWITCHFREQ = 0;
	}
	if (equal(Command, "status") || equal(Command, "switchafter")) {
		format(s, 64, "PTB: (switchafter) Switching starts after %d round(s).", PTB_SWITCHAFTER);
		client_print(id,print_console,s);
	}

	// PTB_SWITCHMIN
	if (equal(Command, "switchmin") && strlen(Argument)) {
		PTB_SWITCHMIN = strtonum(Argument);
		if (PTB_SWITCHMIN < 0) PTB_SWITCHMIN = 0;
	}
	if (equal(Command, "status") || equal(Command, "switchmin")) {
		format(s, 64, "PTB: (switchmin) Switching needs at least %d player(s).", PTB_SWITCHMIN);
		client_print(id,print_console,s);
	}

	// PTB_PLAYERFREQ
	if (equal(Command, "playerfreq") && strlen(Argument)) {
		PTB_PLAYERFREQ = strtonum(Argument);
		if (PTB_PLAYERFREQ < 0) PTB_PLAYERFREQ = 0;
	}
	if (equal(Command, "status") || equal(Command, "playerfreq")) {
		format(s, 64, "PTB: (playerfreq) Individual players are switched every %d round(s) at maximum.", PTB_PLAYERFREQ);
		client_print(id,print_console,s);
	}

	// PTB_SWITCHFREQ
	if (equal(Command, "switchfreq") && strlen(Argument)) {
		PTB_SWITCHFREQ = strtonum(Argument);
		if (PTB_SWITCHFREQ < 1) PTB_SWITCHFREQ = 1;
	}
	if (equal(Command, "status") || equal(Command, "switchfreq")) {
		format(s, 64, "PTB: (switchfreq) Switch occurs every %d round(s) at maximum.", PTB_SWITCHFREQ);
		client_print(id,print_console,s);
	}

	// PTB_FORCESWITCH
	if (equal(Command, "forceswitch") && strlen(Argument)) {
		PTB_FORCESWITCH = strtonum(Argument);
		if (PTB_FORCESWITCH < 0) PTB_FORCESWITCH = 0;
	}
	if (equal(Command, "status") || equal(Command, "forceswitch")) {
		format(ls, 128, "PTB: (forceswitch) Forcing switch after %d unsuccessful switch(es).", PTB_FORCESWITCH);
		client_print(id,print_console,ls);
	}

	// PTB_DEADONLY
	if (equal(Command, "deadonly") && strlen(Argument)) PTB_DEADONLY = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "deadonly")) { if (PTB_DEADONLY) client_print(id,print_console,"PTB: (deadonly) Switching dead only is ON."); else client_print(id,print_console,"PTB: (deadonly) Switching dead only is OFF."); }


	// messages
	if (equal(Command, "status")) client_print(id,print_console,"PTB: ---------- Messages ----------");

	// PTB_TELLWTJ
	if (equal(Command, "tellwtj") && strlen(Argument)) PTB_TELLWTJ = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "tellwtj")) { if (PTB_TELLWTJ) client_print(id,print_console,"PTB: (tellwtj) Telling about WTJ tries is ON."); else client_print(id,print_console,"PTB: (tellwtj) Telling about WTJ tries is OFF."); }

	// PTB_ANNOUNCE
	if (equal(Command, "announce") && strlen(Argument)) PTB_ANNOUNCE = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "announce")) { if (PTB_ANNOUNCE) client_print(id,print_console,"PTB: (announce) Announcements are ON."); else client_print(id,print_console,"PTB: (announce) Announcements are OFF."); }
	
	// PTB_SAYOK
	if (equal(Command, "sayok") && strlen(Argument)) PTB_SAYOK = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "sayok")) { if (PTB_SAYOK) client_print(id,print_console,"PTB: (sayok) ^"OK^" announcements are ON."); else client_print(id,print_console,"PTB: (sayok) ^"OK^" announcements are OFF."); }
	
	// PTB_TYPESAY
	if (equal(Command, "typesay") && strlen(Argument)) PTB_TYPESAY = check_param(Argument);
	if (equal(Command, "status") || equal(Command, "typesay")) { if (PTB_TYPESAY) client_print(id,print_console,"PTB: (typesay) typesay usage is ON."); else client_print(id,print_console,"PTB: (typesay) typesay usage is OFF."); }
	

	// team strength limits
	if (equal(Command, "status")) client_print(id,print_console,"PTB: ---------- Team Strength Limits ----------");

	// PTB_MAXSTREAK
	if (equal(Command, "maxstreak") && strlen(Argument)) {
		PTB_MAXSTREAK = strtonum(Argument);
		if (PTB_MAXSTREAK < 1) PTB_MAXSTREAK = 1;
	}
	if (equal(Command, "status") || equal(Command, "maxstreak")) {
		format(s, 64, "PTB: (maxstreak) Maximum accepted win streak is %d.", PTB_MAXSTREAK);
		client_print(id,print_console,s);
	}

	// PTB_MAXSTREAK
	if (equal(Command, "maxscore") && strlen(Argument)) {
		PTB_MAXSCORE = strtonum(Argument);
		if (PTB_MAXSCORE < 1) PTB_MAXSCORE = 1;
	}
	if (equal(Command, "status") || equal(Command, "maxscore")) {
		format(s, 64, "PTB: (maxscore) Maximum accepted team score difference is %d.", PTB_MAXSCORE);
		client_print(id,print_console,s);
	}

	// PTB_MINRATING
	if (equal(Command, "minrating") && strlen(Argument)) {
    	PTB_MINRATING = fixedstr(Argument);
    	if (PTB_MINRATING < fixed(1)) PTB_MINRATING = fixed(1);
    }
	if (equal(Command, "status") || equal(Command, "minrating")) {
		format(s, 64, "PTB: (minrating) Minimum critical strength rating is %d.%d.",
			fround(PTB_MINRATING, fround_floor),
			fround(fmul(fixed(10), ffract(PTB_MINRATING)), fround_floor)
		);
		client_print(id,print_console,s);
	}

	// PTB_MAXRATING
	if (equal(Command, "maxrating") && strlen(Argument)) {
		PTB_MAXRATING = fixedstr(Argument);
		if (PTB_MAXRATING < fixed(1)) PTB_MAXRATING = fixed(1);
	}
	if (equal(Command, "status") || equal(Command, "maxrating")) {
		format(s, 64, "PTB: (maxrating) Maximum critical strength rating is %d.%d.",
			fround(PTB_MAXRATING, fround_floor),
			fround(fmul(fixed(10), ffract(PTB_MAXRATING)), fround_floor)
		);
		client_print(id,print_console,s);
	}

	// PTB_SUPERRATING
	if (equal(Command, "superrating") && strlen(Argument)) {
		PTB_SUPERRATING = fixedstr(Argument);
		if (PTB_SUPERRATING < fixed(1)) PTB_SUPERRATING = fixed(1);
	}
	if (equal(Command, "status") || equal(Command, "superrating")) {
		format(s, 64, "PTB: (superrating) Super critical strength rating is %d.%d.",
			fround(PTB_SUPERRATING, fround_floor),
			fround(fmul(fixed(10), ffract(PTB_SUPERRATING)), fround_floor)
		);
		client_print(id,print_console,s);
	}

	// misc
	if (equal(Command, "status")) {
		client_print(id,print_console,"PTB: ---------- Misc ----------");
		client_print(id,print_console,"PTB: To enable or disable PTB, type ^"amx_ptb <^"on^"|^"1^"|^"off^"|^"0^">^".");
		client_print(id,print_console,"PTB: To view or change a single PTB setting, type ^"amx_ptb <setting> <^"on^"|^"off^"|value>^".");
		client_print(id,print_console,"PTB: To view a brief overview of PTB commands, type ^"amx_ptb help^" or ^"amx_ptb list^".");
		client_print(id,print_console,"PTB: To save all PTB settings to ^"%s^", type ^"amx_ptb save^".",vaultfile);
		client_print(id,print_console,"PTB: To load all PTB settings from ^"%s^", type ^"amx_ptb load^".",vaultfile);
		client_print(id,print_console,"PTB: For PTB statistics, simply type ^"amx_ptb^".");
	}
    
	return PLUGIN_HANDLED;
}

stock saveSettings() {
	new VAULTDATA[64];

	// team selection control
	if (PTB_LIMITJOIN) set_vaultdata("PTB_LIMITJOIN", "on");
	else set_vaultdata("PTB_LIMITJOIN", "off");
	format(VAULTDATA, 64, "%d", PTB_LIMITAFTER);
	set_vaultdata("PTB_LIMITAFTER", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_LIMITMIN);
	set_vaultdata("PTB_LIMITMIN", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_MAXSIZE);
	set_vaultdata("PTB_MAXSIZE", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_MAXDIFF);
	set_vaultdata("PTB_MAXDIFF", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_AUTOROUNDS);
	set_vaultdata("PTB_AUTOROUNDS", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_WTJAUTO);
	set_vaultdata("PTB_WTJAUTO", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_WTJKICK);
	set_vaultdata("PTB_WTJKICK", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_KICK);
	set_vaultdata("PTB_KICK", VAULTDATA);
	if (PTB_SAVEWTJ) set_vaultdata("PTB_SAVEWTJ", "on");
	else set_vaultdata("PTB_SAVEWTJ", "off");

	// team balancing actions
	if (PTB_SWITCH) set_vaultdata("PTB_SWITCH", "on");
	else set_vaultdata("PTB_SWITCH", "off");
	format(VAULTDATA, 64, "%d", PTB_SWITCHAFTER);
	set_vaultdata("PTB_SWITCHAFTER", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_SWITCHMIN);
	set_vaultdata("PTB_SWITCHMIN", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_SWITCHFREQ);
	set_vaultdata("PTB_SWITCHFREQ", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_PLAYERFREQ);
	set_vaultdata("PTB_PLAYERFREQ", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_FORCESWITCH);
	set_vaultdata("PTB_FORCESWITCH", VAULTDATA);
	if (PTB_DEADONLY) set_vaultdata("PTB_DEADONLY", "on");
	else set_vaultdata("PTB_DEADONLY", "off");

	// messages
	if (PTB_TELLWTJ) set_vaultdata("PTB_TELLWTJ", "on");
	else set_vaultdata("PTB_TELLWTJ", "off");
	if (PTB_ANNOUNCE) set_vaultdata("PTB_ANNOUNCE", "on");
	else set_vaultdata("PTB_ANNOUNCE", "off");
	if (PTB_SAYOK) set_vaultdata("PTB_SAYOK", "on");
	else set_vaultdata("PTB_SAYOK", "off");
	if (PTB_TYPESAY) set_vaultdata("PTB_TYPESAY", "on");
	else set_vaultdata("PTB_TYPESAY", "off");

	// team strength limits
	format(VAULTDATA, 64, "%d", PTB_MAXSTREAK);
	set_vaultdata("PTB_MAXSTREAK", VAULTDATA);
	format(VAULTDATA, 64, "%d", PTB_MAXSCORE);
	set_vaultdata("PTB_MAXSCORE", VAULTDATA);
	format(VAULTDATA, 64, "%d.%d",
		fround(PTB_MINRATING, fround_floor),
		fround(fmul(fixed(10), ffract(PTB_MINRATING)), fround_floor)
	);
	set_vaultdata("PTB_MINRATING", VAULTDATA);
	format(VAULTDATA, 64, "%d.%d",
		fround(PTB_MAXRATING, fround_floor),
		fround(fmul(fixed(10), ffract(PTB_MAXRATING)), fround_floor)
	);
	set_vaultdata("PTB_MAXRATING", VAULTDATA);
	format(VAULTDATA, 64, "%d.%d",
		fround(PTB_SUPERRATING, fround_floor),
		fround(fmul(fixed(10), ffract(PTB_SUPERRATING)), fround_floor)
	);
	set_vaultdata("PTB_SUPERRATING", VAULTDATA);
}

stock loadSettings() {
	new VAULTDATA[64];

	// team selection control
	if (get_vaultdata("PTB_LIMITJOIN", VAULTDATA, 64 ) != 0)
		PTB_LIMITJOIN = check_param(VAULTDATA);
	if (get_vaultdata("PTB_LIMITAFTER", VAULTDATA, 64 ) != 0)
		PTB_LIMITAFTER = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_LIMITMIN", VAULTDATA, 64 ) != 0)
		PTB_LIMITMIN = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_MAXSIZE", VAULTDATA, 64 ) != 0)
		PTB_MAXSIZE = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_MAXDIFF", VAULTDATA, 64 ) != 0)
		PTB_MAXDIFF = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_AUTOROUNDS", VAULTDATA, 64 ) != 0)
		PTB_AUTOROUNDS = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_WTJAUTO", VAULTDATA, 64 ) != 0)
		PTB_WTJAUTO = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_WTJKICK", VAULTDATA, 64 ) != 0)
		PTB_WTJKICK = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_KICK", VAULTDATA, 64 ) != 0)
		PTB_KICK = check_param(VAULTDATA);
	if (get_vaultdata("PTB_SAVEWTJ", VAULTDATA, 64 ) != 0)
		PTB_SAVEWTJ = check_param(VAULTDATA);

	// team balancing actions
	if (get_vaultdata("PTB_SWITCH", VAULTDATA, 64 ) != 0)
		PTB_SWITCH = check_param(VAULTDATA);
	if (get_vaultdata("PTB_SWITCHAFTER", VAULTDATA, 64 ) != 0)
		PTB_SWITCHAFTER = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_SWITCHMIN", VAULTDATA, 64 ) != 0)
		PTB_SWITCHMIN = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_SWITCHFREQ", VAULTDATA, 64 ) != 0)
		PTB_SWITCHFREQ = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_PLAYERFREQ", VAULTDATA, 64 ) != 0)
		PTB_PLAYERFREQ = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_FORCESWITCH", VAULTDATA, 64 ) != 0)
		PTB_FORCESWITCH = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_DEADONLY", VAULTDATA, 64 ) != 0)
		PTB_DEADONLY = check_param(VAULTDATA);

	// messages
	if (get_vaultdata("PTB_TELLWTJ", VAULTDATA, 64 ) != 0)
		PTB_TELLWTJ = check_param(VAULTDATA);
	if (get_vaultdata("PTB_ANNOUNCE", VAULTDATA, 64 ) != 0)
		PTB_ANNOUNCE = check_param(VAULTDATA);
	if (get_vaultdata("PTB_SAYOK", VAULTDATA, 64 ) != 0)
		PTB_SAYOK = check_param(VAULTDATA);
	if (get_vaultdata("PTB_TYPESAY", VAULTDATA, 64 ) != 0)
		PTB_TYPESAY = check_param(VAULTDATA);

	// team strength limits
	if (get_vaultdata("PTB_MAXSTREAK", VAULTDATA, 64 ) != 0)
		PTB_MAXSTREAK = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_MAXSCORE", VAULTDATA, 64 ) != 0)
		PTB_MAXSCORE = strtonum(VAULTDATA);
	if (get_vaultdata("PTB_MINRATING", VAULTDATA, 64 ) != 0)
		PTB_MINRATING = fixedstr(VAULTDATA);
	if (get_vaultdata("PTB_MAXRATING", VAULTDATA, 64 ) != 0)
		PTB_MAXRATING = fixedstr(VAULTDATA);
	if (get_vaultdata("PTB_SUPERRATING", VAULTDATA, 64 ) != 0)
		PTB_SUPERRATING = fixedstr(VAULTDATA);
}

stock displayStatistics(toLog = false, id = 0) {
	//say("displayStatistics");

	new text[64];

	// time
	format(text, 64, "PTB: Statistics generated at: %s", lastTeamBalanceCheck);
	if (toLog) log(text);
	else client_print(id,print_console,text);

	// connected players
	format(text, 64, "PTB: Connected players: %d", get_playersnum());
	if (toLog) log(text);
	else client_print(id,print_console,text);

	// team sizes
	format(text, 64, "PTB: Team sizes: CTs %d, Ts %d", teamCounts[CTS], teamCounts[TS]);
	if (toLog) log(text);
	else client_print(id,print_console,text);

	// team scores
	format(text, 64, "PTB: Team scores: CTs %d, Ts %d", teamScores[CTS], teamScores[TS]);
	if (toLog) log(text);
	else client_print(id,print_console,text);

	// Kills:Deaths
	format(text, 64, "PTB: Team kills:deaths: CTs %d:%d, Ts %d:%d", teamKills[CTS], teamDeaths[CTS], teamKills[TS], teamDeaths[TS]);
	if (toLog) log(text);
	else client_print(id,print_console,text);

	// Kills/Deaths
	format(text, 64, "PTB: Team kills/deaths: CTs %d.%d, Ts %d.%d",
		fround(ctKD, fround_floor),
		fround(fmul(fixed(100), ffract(ctKD)), fround_floor),
		fround(tKD, fround_floor),
		fround(fmul(fixed(100), ffract(tKD)), fround_floor)
	);
	if (toLog) log(text);
	else client_print(id,print_console,text);

	// strength
	format(text, 64, "PTB: Team strengths: CTs %d.%d, Ts %d.%d",
		fround(ctStrength, fround_floor),
		fround(fmul(fixed(100), ffract(ctStrength)),
		fround_floor), fround(tStrength, fround_floor),
		fround(fmul(fixed(100), ffract(tStrength)), fround_floor)
	);
	if (toLog) log(text);
	else client_print(id,print_console,text);

	// rating
	format(text, 64, "PTB: Team ratings: CTs %d.%d, Ts %d.%d",
		fround(ctRating, fround_floor),
		fround(fmul(fixed(100), ffract(ctRating)), fround_floor),
		fround(tRating, fround_floor),
		fround(fmul(fixed(100), ffract(tRating)), fround_floor)
	);
	if (toLog) log(text);
	else client_print(id,print_console,text);

	// won rounds
	if (winStreaks[CTS] > 0) {
		format(text, 64, "PTB: Last %d round(s) won by CTs.", winStreaks[CTS]);
		if (toLog) log(text);
		else client_print(id,print_console,text);
	}
	else if (winStreaks[TS] > 0) {
		format(text, 64, "PTB: Last %d round(s) won by Ts.", winStreaks[TS]);
		if (toLog) log(text);
		else client_print(id,print_console,text);
	}
	
	// winning team
	if (winnerTeam == CTS) format(text, 64, "PTB: The CTs are the winning team.");
	else if (winnerTeam == TS) format(text, 64, "PTB: The Ts are the winning team.");
	else format(text, 64, "PTB: Teams are balanced.");
	if (toLog) log(text);
	else client_print(id,print_console,text);

	format(text, 64, "PTB: (These statistics might be already outdated.)");
	if (toLog) log(text);
	else client_print(id,print_console,text);
	new ltext[80]
	format(ltext, 80, "PTB: To view a brief overview of PTB commands, type ^"amx_ptb help^" or ^"amx_ptb list^".");
	if (toLog) log(text);
	else client_print(id,print_console,text);

	format(text, 64, "PTB: To view all PTB settings, type ^"amx_ptb status^".");
	if (toLog) log(text);
	else client_print(id,print_console,text);
}

public client_connect(id)
{
	//say("plugin_connect");

	if ((id >= 1) && (id <= 32)) {
		kills[id] = 0;
		deaths[id] = 0;
		isChoosingTeam[id] = 1;
		isBeingTransfered[id] = 0;
		playerTeam[id] = 0;
		lastRoundSwitched[id] = 0;
		wtjCount[id] = 0;
	}
	else {
		say("PTB: Shouldn't get here, unexpected id.");
	}
	return PLUGIN_CONTINUE;
}

public client_disconnect(id)
{
	//say("plugin_disconnect");

	//new debug[64];
	//format(debug, 64, "[DEBUG] PTB plugin_disconnect: %d %d", UserIndex, playerTeam[UserIndex]);
	//log(debug);
	
	if ((id >= 1) && (id <= 32)) {
		kills[id] = 0;
		deaths[id] = 0;
		isChoosingTeam[id] = 0;
		isBeingTransfered[id] = 0;
		if (playerTeam[id] == TS || playerTeam[id] == CTS) {
			teamCounts[playerTeam[id]] -= 1;
		}
		playerTeam[id] = 0;
		lastRoundSwitched[id] = 0;
		wtjCount[id] = 0;
	}
	else {
		say("PTB: Shouldn't get here, unexpected id.");
	}
	
	// redundant team size check
	teamCounts[CTS] = 0;
	teamCounts[TS] = 0;
	teamCounts[0] = 0;
	for (new i = 1; i <= 32; ++i) teamCounts[playerTeam[i]] += 1;

	//displayStatistics(true);
	
	return PLUGIN_CONTINUE;
}




public plugin_init()
{
	//say("plugin_init");

	//initVariables();
	copy(vaultfile,64,"addons/amx/ptb.ini")
	// must du this here
	PTB_MINRATING = fixedstr(MIN_RATING);
	PTB_MAXRATING = fixedstr(MAX_RATING);
	PTB_SUPERRATING = fixedstr(SUPER_RATING);

	register_plugin("Ptahhotep's Team Balancer", "1.6","Ptahhotep");
	register_clcmd("amx_ptb", "amx_ptb", ACCESS_ALL, "amx_ptb <^"on^" | ^"off^" | ^"status^" | ^"^">: Switch Ptahhotep's Team Balancer on or off, get status or statistics.");

	//register_clcmd("admin_transfer", "admin_transfer", ACCESS_BALANCE);
	//register_clcmd("admin_spectator", "admin_spectator", ACCESS_BALANCE);

	register_clcmd("chooseteam", "ChoosingTeam", ACCESS_ALL);
	register_clcmd("menuselect", "MenuSelect", ACCESS_ALL);
	register_clcmd("radio", "NotChoosingTeam", ACCESS_ALL);
	register_clcmd("buy", "NotChoosingTeam", ACCESS_ALL);
	register_clcmd("buyequip", "NotChoosingTeam", ACCESS_ALL);
	register_clcmd("showbriefing", "NotChoosingTeam", ACCESS_ALL);
/*
	register_srvcmd("ptb_teamselection", "ptb_teamselection");
	register_srvcmd("ptb_kill", "ptb_kill");
	register_srvcmd("ptb_teamaction", "ptb_teamaction");
	register_srvcmd("ptb_worldaction", "ptb_worldaction");
	
	exec("logd_reg 54 ptb_teamselection");
	exec("logd_reg 57 ptb_kill");
	exec("logd_reg 61 ptb_teamaction");
	exec("logd_reg 62 ptb_worldaction");
*/
	loadSettings();
	logd_init()
	return PLUGIN_CONTINUE;
}
